﻿---------------------------------------------
-- StunWatch v2.2 BETA2
-- Changes:
----Aceified
----Fixed reapplication of stuns
----Scale feature
----Diminishing Returns
----Selective Stun Tracking
-- TODO:
----Switch to defaults option in slash menu
----Fix twitchy checkboxes in Waterfall config GUI
----if not lib then error return end // to protect against missing libs
---------------------------------------------

StunWatch = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0")

local UnitName = UnitName
local UnitGUID = UnitGUID
local StunWatch = StunWatch
--We'll assign this later
local db
--This, too
local sdb = {}
--For tracking GUIDs so that we can unregister all candybars after a unit dies
local unitDB = {}
--For sap
local sapID
--Frame, identifier, and libs
local StunWatchFrame
local StunWatchName = "StunWatchBars"
local candy = LibStub:GetLibrary("LibCandyBar")
local waterfall = AceLibrary:HasInstance("Waterfall-1.0") and AceLibrary("Waterfall-1.0") or nil

--Remove a candybar based on the ID given
function StunWatch:KillTimer(id)
	candy:StopCandyBar(id)
	candy:UnregisterCandyBar(id)
	--Get the spell ID and unit GUID from the ID we were given
	local spellName, unitGUID = strsplit(":", id)
	--Remove this timer from the database
	unitDB[unitGUID][spellName] = nil
end

--Remove ALL timers from the specified unit
function StunWatch:KillAllTimers(unitGUID)
	--terate through the table removing all candybars associated with the unit
	for k, v in pairs(unitDB[unitGUID]) do
		candy:UnregisterCandyBar(v)
		unitDB[unitGUID][k] = nil
	end
	unitDB[unitGUID] = nil
end

do
	--Check for sap, this way, we can delete any other sap timers if we happen to be out of range for SPELL_AURA_REMOVED
	local function sap(id)
		if (spellName == sdb[6770]) then
			if (sapID) then
				if candy:IsCandyBarRegistered(sapID) then
					candy:UnregisterCandyBar(sapID)
				end
			end
			sapID = id
		end
	end
	
	function StunWatch:StartTimer(spellName, unitGUID, unitName, duration)
		--Create a unique ID based on the spell name/unit GUID
		local id = (spellName .. ":" .. unitGUID)
		sap(id)
		--Check if we have a database on the current unit
		if (not unitDB[unitGUID]) then
			--if not, create one with the specified spell
			unitDB[unitGUID] = { [spellName] = id }
		else
			--if there is, add the spell to the DB
			unitDB[unitGUID][spellName] = id
		end
		--Create the text of the bar based on the spell and unit names
		local text = (spellName .. ": " .. unitName)
		local invert = (tostring(db.invert))
		candy:RegisterCandyBar(id, duration, text, nil, 1, 1, 0)
		candy:SetCandyBarReversed(id, invert)
		candy:RegisterCandyBarWithGroup(id, StunWatchName)
		candy:SetWidth(id, db.width) --было 160
		candy:SetHeight(id, 10)
		candy:SetCandyBarScale(id, db.scale)
		candy:StartCandyBar(id, true)
		self:TargetFocusUpdate();
	end

	function StunWatch:RetainDebuff(spellName, unitGUID, unitName, duration, maxDuration)
		--This way, we can start a new timer without screwing up existing ones
		local id = (spellName .. ":" .. unitGUID)
		local invert = (tostring(db.invert))
		candy:SetCandyBarReversed(id, invert)
		candy:SetCandyBarTime(id, maxDuration)
		candy:SetCandyBarTimeLeft(id, duration)
	end
end

function StunWatch:CreateMainFrame()
	--Create and edit the frame which the user uses to move the candybar group
	local frame = StunWatchFrame
	frame:SetWidth("300") --170
	frame:SetHeight("20")
	frame:SetBackdrop({
		bgFile = "Interface\\Tooltips\\UI-Tooltip-Background",
		tile = true,
		tileSize = 16,
		edgeSize = 0,
		insets = { left = 0, right = 0, top = 0, bottom = 0 },
	})
	frame:SetMovable(true)
	frame:SetClampedToScreen(true)
	frame:SetPoint("CENTER", UIParent, "CENTER", 0 ,0)
	frame.Text = frame:CreateFontString(nil, "OVERLAY")
	frame.Text:SetFontObject("GameFontNormal")
	frame.Text:SetWidth(300) -- было 170
	frame.Text:SetHeight(15)
	frame.Text:SetPoint("CENTER")
	frame.Text:SetText(StunWatchName)
	--SetScripts
	frame:SetScript("OnMouseDown", function(self, button) if (button == "LeftButton") then self:StartMoving(); end end)
	frame:SetScript("OnMouseUp", function(self, button) if (button == "LeftButton") then self:StopMovingOrSizing(); end end)
end

function StunWatch:Initialize()
	local select = select
	local sdb = {
		-- Cheap Shot
		[1833] = select(1, GetSpellInfo(1833)),
		-- Kidney Shot
		[408] = select(1, GetSpellInfo(408)),
		-- Blind
		[2094] = select(1, GetSpellInfo(2094)),
		-- Gouge
		[38764] = select(1, GetSpellInfo(38764)),
		-- Sap
		[6770] = select(1, GetSpellInfo(6770)),
		-- Expose Armor
		[8647] = select(1, GetSpellInfo(8647)),
		-- Rupture
		[1943] = select(1, GetSpellInfo(1943)),
		--  Garrote
		[703] = select(1, GetSpellInfo(703)),
		-- Mace Stun
		[5530] = select(1, GetSpellInfo(5530)),
		-- Blade Twisting
		[31125] = select(1, GetSpellInfo(31125)),
		--Deadly Throw
		[26679] = select(1, GetSpellInfo(26679)),
		--Imp. Kick
		[18425] = select(1, GetSpellInfo(18425)),
		--Riposte
		[14251] = select(1, GetSpellInfo(14251)),
	}
	--set up our database
	self.defaults = {
		lock = true,
		grow = true,
		invert = 1,
		target = true,
		focus = true,
		scale = 1,
		width = 150,
		sdbn = {},
	}
	--Set up spell name database, for checking against spells in the UNIT_AURA event, and various other places
	for k,v in pairs(sdb) do
		self.defaults.sdbn[v] = 1
	end
	self:RegisterDB("StunWatchDB")
	self:RegisterDefaults("profile", self.defaults)
	db = self.db.profile
	
	local frame = StunWatchFrame
	--set up our options
	self.opts = {
		type ='group',
		name = "StunWatch",
		args = {
			lock = {
				type = 'toggle',
				name = "Закрепить",
				desc = "Закрепить StunWatch",
				order = 2,
				get = function() return db.lock end,
				set = function(v)	
					db.lock = v
					if (not db.lock) then
						frame:EnableMouse(true)
						frame:Show()
					else
						frame:EnableMouse(false)
						frame:Hide()
					end
					
				end,
			},
			invert = {
				type = 'toggle',
				name = "Инверсия",
				desc = "Инверсия заполнения полос",
				order = 3,
				get = function() return db.invert end,
				set = function(v)
					if (not v) then
						db.invert = false
					else
						db.invert = "1"
					end
				end,
			},
			grow = {
				type = 'toggle',
				name = "Рост вверх",
				desc = "Переключение роста вверх и вниз",
				order = 4,
				get = function() return db.grow end,
				set = function(v)
					if (v) then
						candy:SetCandyBarGroupGrowth(StunWatchName, true) -- bars grow upwards
					else
						candy:SetCandyBarGroupGrowth(StunWatchName, false) --bars grow downwards
					end
					
					db.grow = v
				end,
			},
			target = {
				type = 'toggle',
				name = "Цель (рекомендуется)",
				desc = "Отслеживать Оглушения/Дебаффов на вашей цели",
				order = 5,
				get = function() return db.target end,
				set = function(v)
					if (not v) then
						db.target = false
						self:UnregisterEvent("PLAYER_TARGET_CHANGED", "TargetFocusUpdate")
					else
						db.target = true
						self:RegisterEvent("PLAYER_TARGET_CHANGED", "TargetFocusUpdate")
					end
				end,
			},
			focus = {
				type = 'toggle',
				name = "Фокус",
				desc = "Отслеживать Оглушения/Дебаффов на вашем фокусе",
				order = 6,
				get = function() return db.focus end,
				set = function(v)
					if (not v) then
						db.focus = false
						self:UnregisterEvent("PLAYER_FOCUS_CHANGED", "TargetFocusUpdate")
					else
						db.focus = true
						self:RegisterEvent("PLAYER_FOCUS_CHANGED", "TargetFocusUpdate")
					end
				end,
			},
			scale = {
				type = 'range',
				name = "Масштаб",
				desc = "Масштаб полос StunWatch",
				order = 7,
				get = function() return db.scale end,
				set = function(v)
					db.scale = v
					frame:SetScale(v)
				end,
				min = .5,
				max = 1.5,
				step = .1,
			},
			width = {
				type = 'range',
				name = "Ширина",
				desc = "Ширина полос StunWatch",
				order = 8,
				get = function() return db.width end,
				set = function(v)
					db.width = v
					frame:SetWidth(v)
				end,
				min = 150,
				max = 300,
				step = .1,
			},			
			debuffFilter = {
				type = 'group',
				name = "Дебаффы",
				desc = "Выберите какой дебафф отслеживать",
				order = 9,
				args = {
					cheapshot = {
						type = 'toggle',
						name = sdb[1833],
						desc = "Вкл/выкл: " .. sdb[1833],
						order = 10,
						get = function() return db.sdbn[sdb[1833]] end,
						set = function(v)
							if (not v) then
								db.sdbn[sdb[1833]] = false
							else
								db.sdbn[sdb[1833]] = 1
							end
						end,
					},
					kidneyshot = {
						type = 'toggle',
						name = sdb[408],
						desc = "Вкл/выкл: " .. sdb[408],
						order = 11,
						get = function() return db.sdbn[sdb[408]] end,
						set = function(v)
							if (not v) then
								db.sdbn[sdb[408]] = false
							else
								db.sdbn[sdb[408]] = 1
							end
						end,
					},
					blind = {
						type = 'toggle',
						name = sdb[2094],
						desc = "Вкл/выкл: " .. sdb[2094],
						order = 12,
						get = function() return db.sdbn[sdb[2094]] end,
						set = function(v)
							if (not v) then
								db.sdbn[sdb[2094]] = false
							else
								db.sdbn[sdb[2094]] = 1
							end
						end,
					},
					gouge = {
						type = 'toggle',
						name = sdb[38764],
						desc = "Вкл/выкл: " .. sdb[38764],
						order = 13,
						get = function() return db.sdbn[sdb[38764]] end,
						set = function(v)
							if (not v) then
								db.sdbn[sdb[38764]] = false
							else
								db.sdbn[sdb[38764]] = 1
							end
						end,
					},
					sap = {
						type = 'toggle',
						name = sdb[6770],
						desc = "Вкл/выкл: " .. sdb[6770],
						order = 14,
						get = function() return db.sdbn[sdb[6770]] end,
						set = function(v)
							if (not v) then
								db.sdbn[sdb[6770]] = false
							else
								db.sdbn[sdb[6770]] = 1
							end
						end,
					},
					exposearmor = {
						type = 'toggle',
						name = sdb[8647],
						desc = "Вкл/выкл: " .. sdb[8647],
						order = 15,
						get = function() return db.sdbn[sdb[8647]] end,
						set = function(v)
							if (not v) then
								db.sdbn[sdb[8647]] = false
							else
								db.sdbn[sdb[8647]] = 1
							end
						end,
					},
					rupture = {
						type = 'toggle',
						name = sdb[1943],
						desc = "Вкл/выкл: " .. sdb[1943],
						order = 16,
						get = function() return db.sdbn[sdb[1943]] end,
						set = function(v)
							if (not v) then
								db.sdbn[sdb[1943]] = false
							else
								db.sdbn[sdb[1943]] = 1
							end
						end,
					},
					garrote = {
						type = 'toggle',
						name = sdb[703],
						desc = "Вкл/выкл: " .. sdb[703],
						order = 17,
						get = function() return db.sdbn[sdb[703]] end,
						set = function(v)
							if (not v) then
								db.sdbn[sdb[703]] = false
							else
								db.sdbn[sdb[703]] = 1
							end
						end,
					},
					macestun = {
						type = 'toggle',
						name = sdb[5530],
						desc = "Вкл/выкл: " .. sdb[5530],
						order = 18,
						get = function() return db.sdbn[sdb[5530]] end,
						set = function(v)
							if (not v) then
								db.sdbn[sdb[5530]] = false
							else
								db.sdbn[sdb[5530]] = 1
							end
						end,
					},
					bladetwisting = {
						type = 'toggle',
						name = sdb[31125],
						desc = "Вкл/выкл: " .. sdb[31125],
						order = 19,
						get = function() return db.sdbn[sdb[31125]] end,
						set = function(v)
							if (not v) then
								db.sdbn[sdb[31125]] = false
							else
								db.sdbn[sdb[31125]] = 1
							end
						end,
					},
					deadlythrow = {
						type = 'toggle',
						name = sdb[26679],
						desc = "Вкл/выкл: " .. sdb[26679],
						order = 20,
						get = function() return db.sdbn[sdb[26679]] end,
						set = function(v)
							if (not v) then
								db.sdbn[sdb[26679]] = false
							else
								db.sdbn[sdb[26679]] = 1
							end
						end,
					},
					impkick = {
						type = 'toggle',
						name = sdb[18425],
						desc = "Вкл/выкл: " .. sdb[18425],
						order = 21,
						get = function() return db.sdbn[sdb[18425]] end,
						set = function(v)
							if (not v) then
								db.sdbn[sdb[18425]] = false
							else
								db.sdbn[sdb[18425]] = 1
							end
						end,
					},
					riposte = {
						type = 'toggle',
						name = sdb[14251],
						desc = "Вкл/выкл: " .. sdb[14251],
						order = 22,
						get = function() return db.sdbn[sdb[14251]] end,
						set = function(v)
							if (not v) then
								db.sdbn[sdb[14251]] = false
							else
								db.sdbn[sdb[14251]] = 1
							end
						end,
					},
				},
			},
		},
	}
	--Set up our slash command, depending upon waterfall's availibility
	if (waterfall) then
		waterfall:Register("StunWatch", "aceOptions", self.opts, "colorR", 0.2, "colorG", 0.3, "colorB", 0.6)
		self:RegisterChatCommand("/sw", "/stunwatch", function() waterfall:Open("StunWatch") end)
	else
		self:RegisterChatCommand("/sw", "/stunwatch", self.opts)
	end
	
	if (not db.lock) then 
		frame:EnableMouse(true)
		frame:Show() 
	else 
		frame:Hide()
		frame:EnableMouse(false)
	end
	local name = StunWatchName
	if (db.grow) then 
		candy:SetCandyBarGroupGrowth(name, true) --bars grow upwards /please test
	else
		candy:SetCandyBarGroupGrowth(name, false) --bars grow downwards /please test
	end
	if (not db.lock) then
		frame:Show()
	else
		frame:Hide()
	end
	frame:SetScale(db.scale)
	candy:RegisterCandyBarGroup(name)
	candy:SetCandyBarGroupPoint(name, "BOTTOM", frame, "BOTTOM", 0, 0)
end


function StunWatch:UpdateBars(token)
	for i=1,40 do
		spellName, _, _, _, _, maxDuration, timeLeft = UnitDebuff(token, i)
		if (not spellName) then
			return
		end
		--Check if it's a spell we want to track
		if (db.sdbn[spellName] and timeLeft and timeLeft > 0) then
			--We need these variables to check for and start timers
			local GUID, name = UnitGUID(token), UnitName(token)
			if (not candy:IsCandyBarRegistered(spellName .. ":" .. GUID)) then
				self:StartTimer(spellName, GUID, name, timeLeft)
			else
				self:RetainDebuff(spellName, GUID, name, timeLeft, maxDuration)
			end
		end
	end
end

function StunWatch:UNIT_AURA(token)
	if (db[token]) then
		self:UpdateBars(token)
	end
end

function StunWatch:COMBAT_LOG_EVENT_UNFILTERED(...)
	--Retrieve the specific combat event
	local _, cmbEvent = select(1, ...)
	if (cmbEvent == "UNIT_DIED") then
		--Check if there are any timers for the unit that died, if there are, clear them all
		local dstGUID = select(6, ...)
		if (unitDB[dstGUID]) then
			self:KillAllTimers(dstGUID)
		end
	elseif ((cmbEvent == "SPELL_AURA_REMOVED") or (cmbEvent == "SPELL_AURA_DISPELLED")) then
		--Clear any debuff timers associated with the removed debuff
		local dstGUID, _, _, _, spellName = select(6, ...)
		local id = (spellName .. ":" .. dstGUID)
		if candy:IsCandyBarRegistered(id) then
			self:KillTimer(id)
		end
	end
end

function StunWatch:TargetFocusUpdate()
	self:UpdateBars("target")
	self:UpdateBars("focus")
end

function StunWatch:OnInitialize()
	self:CreateMainFrame()
	self:Initialize()
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self:RegisterEvent("PLAYER_TARGET_CHANGED", "TargetFocusUpdate")
	self:RegisterEvent("PLAYER_FOCUS_CHANGED", "TargetFocusUpdate")
	self:RegisterEvent("UNIT_AURA")
end

StunWatchFrame = CreateFrame("Frame", "StunWatchMainFrame", UIParent)